"use strict";

var _           = require("underscore");
var error       = require("./error");
var Fs          = require("fs");
var URL         = require("url");
var Path        = require("path");
var Http        = require("http");
var Https       = require("https");
var Querystring = require("querystring");
var Crypto      = require("crypto");
var XML         = require("libxml");

/**
 * A class to call the Heroku REST API. You might use this if you want to
 * manage your Heroku apps from within a Javascript engine, such as Node.js.
 *
 * `HerokuClient` maps specific API calls to methods, and will handle both
 * `xml` or `json` results in response data before returning a callback.
 *
 * There's a method mapped to each HTTP Verb (used in the API to facilitate things)
 * Each request is prepared in the method `request` and responses are passed
 * to `handleHttpError` and `extractWarning` which will handle expectations
 * and API warnings respectively.
 *
 * Static methods:
 *
 *  - `auth` automatically creates an instance and authenticates the user returning
 *    an API key if successful.
 *  - `getLocalFileCaCert` reads data/cacert.pem file, extracts certificates and
 *    merges them back together wrapped between one key BEGIN/END tokens.
 *    The resulting `blob` is passed to `HerokuClient::certificate`.
 *
 * Example:
 *
 *   var heroku = require("node-heroku");
 *   var HerokuClient = new heroku("me@example.com", "mypass");
 *   var herokuClient.create("myapp");
 */
 
var DEFAULT_HOST = "heroku.com";

var HerokuClient = module.exports = function(user, password, host) {
    this.user = user;
    this.password = password;
    this.host = host || DEFAULT_HOST;
};

/**
 * For API reference and to know what each API means, please refer to the CLI
 * docs @link http://devcenter.heroku.com/articles/using-the-cli
 */
(function() {
    
    var jsonHeader = {"Accept": "application/json"};
    var xmlHeader  = {"Accept": "application/xml"};
    
    this.VERSION = require("./constants").VERSION;
    
    //this.RUBY_VERSION = require("./constants").RUBY_VERSION;
    
    this.HEROKU_HEADERS = {
      "X-Heroku-API-Version" : "2",
      "User-Agent"           : "heroku-gem/" + this.VERSION,
      //"X-Ruby-Version"       : process.env.RUBY_VERSION || this.RUBY_VERSION,
      //"X-Ruby-Platform"      : process.env.RUBY_PLATFORM || process.platform
    };
    
    this.login = function(user, password, callback) {
        var credentials = {
            username: user,
            password: password
        };
        
        this.post("/login", credentials, jsonHeader, function(err, data, status) {
            if (err)
                return callback(err);
            
            var apiKey;
            try {
                apiKey = JSON.parse(data).api_key;
            } catch(e) {
                return callback(e);
            }
            
            callback(null, apiKey, status);
        });
    };
    
    // GETs a list of apps which the user is a collaborator on.
    this.list = function(callback) {
        this.get("/apps", jsonHeader, function(err, data) {
            if (err)
                return callback(err);
            
            var apps;
            try {
                apps = JSON.parse(data);
            } catch(e) {
                return callback(e);
            }
            
            if (!Array.isArray(apps))
                return callback(new error.Type("Error retrieving heroku list." +
                    "An array with apps was expected"));
            
            callback(null, apps);
        });
    };
    
    // Show info such as mode, custom domain, and collaborators on an app.
    this.appInfo = function(nameOrDomain, callback) {
        nameOrDomain = nameOrDomain.replace(/^(http:\/\/)?(www\.)?/, "");
        var url = "/apps/" + nameOrDomain;
        
        this.get(url, xmlHeader, function(err, xmlData) {
            if (err)
                return callback(err);

            var doc   = XML.parseFromString(xmlData);
            var nodes = doc.documentElement.selectNodes("*");
            var data  = {};

            for (var i=0; i < nodes.length; i++)
                data[nodes[i].nodeName] = nodes[i].nodeValue;
            
            callback(null, data);
        });
    };
    
    // Create a new app, with an optional name.
    this.create = function(name, options, callback) {
        if (name)
            options.name = name;
        
        var qs = {};
        _.each(options, function(value, key) {
            qs["app["+key+"]"] = value;
        });
        
        this.post("/apps", qs, jsonHeader, function(err, data) {
            if (err)
                return callback(err);
            
            try {
                data = JSON.parse(data);
            } catch(e) {
                return callback(e);
            }
            
            callback(null, data);
        });
    };
    
    this.createComplete = function(name, callback) {
        var url = "/apps/" + name + "/status";
        this.put(url, null, xmlHeader, function(err, xmlData, status) {
            if (err)
                return callback(err);
            
            var doc   = XML.parseFromString(xmlData);
            var nodes = doc.documentElement.selectNodes("*");
            var data  = {};

            for (var i=0; i < nodes.length; i++)
                data[nodes[i].nodeName] = nodes[i].nodeValue;
            
            callback(null, status === 201, data);
        });
    };
    
    this.installAddon = function(addon, callback) {
        throw new Error("Not implemented");
    };
    
    /**
     * Update an app. Available attributes:
     * name: rename the app (changes http and git urls)
     */
    this.update = function(name, attributes, callback) {
        throw new Error("Not implemented");
    };
    
    this.destroy = function(name, callback) {
        var url = "/apps/" + name;
        this.delete(url, jsonHeader, callback);
    };
    
    // Get the list of ssh public keys for the current user.
    this.keys = function(callback) {
        this.get("/user/keys", xmlHeader, function(err, xmlData) {
            var doc     = XML.parseFromString(xmlData);
            var elem    = doc.documentElement;
            var keys    = elem.selectNodes("key/contents");
            var emails  = elem.selectNodes("key/email");
            var data    = [];
            var keyPair = {};
            
            for (var i=0; i < keys.length; i++) {
                keyPair[emails[i].nodeValue] = keys[i].nodeValue.trim();
                data.push(keyPair);
                keyPair = {};
            }
            
            keyPair = null;
            callback(null, data);
        });
    };
    
    // Add an ssh public key to the current user.
    this.addKey = function(key, callback) {
        var headers = _.extend(xmlHeader, {"Content-Type": "text/ssh-authkey"});
        this.post("/user/keys", key, headers, callback);
    };
    
    // Remove an existing ssh public key from the current user.
    this.removeKey = function(key, callback) {
        var url = "/user/keys/" + encodeURIComponent(key).replace(".", "%2E");
        this.delete(url, jsonHeader, callback);
    };
    
    // Clear all keys on the current user.
    this.clearKeys = function(callback) {
        this.delete("/user/keys/", jsonHeader, callback);
    };
    
    /******************/
    
    this.request = function(method, uri, callback) {
        var auth = "Basic " + new Buffer(this.user + ':' + this.password).toString("base64");
        var fullUri = this.parseFullUri(uri);

        var options = {
            host: fullUri.host,
            path: fullUri.href,
            method: method
        };
        if (fullUri.port)
            options.port = fullUri.port;
        
        if (this.certificate())
            options.ca = this.certificate();
        
        var request = (fullUri.protocol === "https:" ? Https : Http).request(options, callback);
        request.setHeader("Authorization", auth);
        
        return request;
    };
    
    this.get = function(uri, extraHeaders, callback) {
        this.process("get", uri, extraHeaders, null, callback);
    };
    
    this.post = function(uri, payload, extraHeaders, callback) {
        this.process("post", uri, extraHeaders, payload, callback);
    };
    
    this.put = function(uri, payload, extraHeaders, callback) {
        this.process("put", uri, extraHeaders, payload, callback);
    };
    
    this["delete"] = function(uri, extraHeaders, callback) {
        this.process("delete", uri, extraHeaders, null, callback);
    };
    
    this.process = function(method, uri, extraHeaders, payload, callback) {
        var self = this;

        if (self[method] === undefined)
            throw new error.InvalidArgument("HTTP method not defined");
        else if (!uri)
            throw new error.InvalidArgument("A valid parameter 'uri' is required");
        
        var headers = {};
        _.extend(headers, this.HEROKU_HEADERS);
        _.extend(headers, extraHeaders || {});
        
        var httpError = false;
        var request = this.request(method, uri, function(response) {
            var data = "";
            
            response.on("data", function(chunk) {
                data += chunk.toString();
            });
            response.on("end", function() {
                if (!httpError)
                    callback(null, data, response.statusCode, response.headers);
                else {
                    var body = self.extractError(data, headers["Accept"]);
                    if (body)
                        httpError.message += " | " + body;
                        
                    callback(httpError);
                }
            });
            response.on("error", function(err) {
                callback(err);
            });
            response.on("httperror", function(err) {
                httpError = err;
            });
            
            self.extractWarning(response);
            self.handleHttpError(response);
        });
        
        _.each(headers, function(value, name) {
            request.setHeader(name, value);
        });
        
        if (payload) {
            var body = (typeof payload === "object")
                ? Querystring.stringify(payload)
                : payload;
            
            request.setHeader("Content-Length", body.length);
            request.end(body);
        } else {
            request.setHeader("Content-Length", 0);
            request.end();
        }
    };
    
    this.handleHttpError = function(response) {
        var code = response.statusCode;
        var m = "";
        
        if (!response.client.authorized && code === 200) {
            code = 403;
            m = "Client Certificate Untrusted or Invalid";
        }
        if (code < 300)
            return; // success
        
        var e;
        if (code > 400 && code <= 505)
            e = new error[code](m);
        else
            e = new error.Http(m, code);
        
        response.emit("httperror", e);
    };
    
    this.parseFullUri = function(given) {
        var fullHost = (/^http/.test(this.host)) ? this.host : "https://api." + this.host;
        var host = URL.parse(fullHost);
        var uri = URL.parse(given);
        
        !uri.protocol && (uri.protocol = host.protocol || "https:");
        !uri.host && (uri.host = host.host);
        (/\:\d+/).test(fullHost) && (uri.port = host.port);
        uri.href = (uri.href.charAt(0) === "/") ? uri.href : "/" + uri.href;
        
        return uri;
    };
    
    this.extractError = function(data, format) {
        var errorStr = "";
        
        if (format === jsonHeader["Accept"]) {
            try {
                data = JSON.parse(data);
                // Suppose it gets the same structure as XML...
                if (data.errors.error)
                    errorStr = data.errors.error;
            } catch(e) {
                errorStr = "" + data;
            }
        } else if (format === xmlHeader["Accept"]) {
            try {
                var doc   = XML.parseFromString(data);
                var nodes = doc.documentElement.selectNodes("error");
                if (typeof nodes[0].nodeValue === "string")
                    errorStr  = nodes[0].nodeValue;
            } catch(e) {
                errorStr = "" + data;
            }
        }
        
        return errorStr;
    };
    
    this.extractWarning = function(response) {
        if (!response)
            return;
        
        if (response.headers["x_heroku_warning"] && this.warningCb) {
            var warning = response.headers["x_heroku_warning"];
            if (!this.displayedWarnings)
                this.displayedWarnings = {};
            
            if (!this.displayedWarnings[warning]) {
                this.warningCb.call(warning);
                this.displayedWarnings[warning] = true;
            }
        }
    };
    
    this.onWarning = function(fn) {
        this.warningCb = fn;
    };
    
    this.certificate = function(pem) {
        return this.cacert || (this.cacert = pem);
    };
    
}).call(HerokuClient.prototype);

HerokuClient.auth = function(user, password, host, callback) {
    var client = new(this)(user, password, host || DEFAULT_HOST);
    
    client.login(user, password, callback);
};

HerokuClient.getLocalFileCaCert = function(file, callback) {
    file = file || Path.resolve(__dirname + "/../../data/cacert.pem");
    
    Fs.readFile(file, "ascii", function(err, contents) {
        if (err)
            throw err;
        
        if (!(/(?:-+BEGIN CERTIFICATE-+)/.test(contents))) {
            callback(contents);
        }
        else {
            var certs = contents.split(/(?:-+BEGIN CERTIFICATE-+)/)
            certs.shift();
            var base64;
            callback(certs.map(function(blob) {
                base64 = blob.split(/(?:-+END CERTIFICATE-+)/)[0];
                return "-----BEGIN CERTIFICATE-----" + base64 + "-----END CERTIFICATE-----";
            }));
        }
    });
};