"use strict";

var EventEmitter = require("events").EventEmitter;
var Sys          = require("sys");
var Path         = require("path");
var Fs           = require("fs");
var Async        = require("asyncjs");
var error        = require("./error");
var _            = require("underscore");
var HerokuClient = require("./client");

/**
 * This class is used to make complex requests to the API, often sending more
 * than one request to get/set/arrange data before returning a callback.
 * It requires a client instance to work with the API methods.
 * `HerokuCommand` also integrates parameter checking and error handling.
 *
 * Every command must contain a protected property $options with mapping values
 * ranging between `true` and `false`.
 * Also, commands are executed using HerokuCommand::run and sensitive syntax,
 * e.g.: `apps:list:`m `keys:add`
 */
var HerokuCommand = module.exports = function(client) {
    if (!client || !(client instanceof HerokuClient)) {
        throw new error.InvalidArgument("First argument passed must be an " +
            "instance of HerokuClient");
    }
    
    EventEmitter.call(this);
    this.currentCommand = null;
    this.$client = client;
    this._commands = {};
    this._load();
};

Sys.inherits(HerokuCommand, EventEmitter);

(function() {
    
    this._load = function() {
        var dir = Path.join(__dirname, "/command");
        var self = this;
        var name = "";
        
        Async.readdir(dir).stat().each(function(file, next) {
            if (!/\.js$/.test(file.name))
                return;

            name = file.name.replace(/\.js$/, "");
            self.registerCommand(name, file.path);
            next();
        }).end(function() {
            self.emit("ready");
        });
    };
    
    this.registerCommand = function(command, file, end) {
        this._commands[command] = require(file);
        this._commands[command].prototype.$client = this.$client;
    };
    
    this.commands = function() {
        return this._commands;
    };
    
    this.prepareRun = function(cmd, args) {
        var command = cmd.split(":");
        if (command.length !== 2) {
            throw new error.InvalidArgument("Error executing the command. " +
                "This command name is incorrect");
        }
        
        if (typeof args !== "object" || _.isArray(args) || _.isFunction(args)) {
            throw new error.Type("Second argument must be an object containing " +
                "key:value pair command arguments");
        }
        
        var object = command[0];
        var method = command[1];

        if ((object in this.commands()) === false)
            throw new error.BadFunctionCall("Command class '" + object + "' does not exists");
        
        var instance = this.commands()[object];
        
        if ((method in instance.prototype) === false) {
            throw new error.InvalidArgument("'" + object + ":" + method +"'" +
                " is not a heroku command");
        }
        
        var opts = {};
        if (args) {
            _.each(args, function(option, name) {
                if (globalOptions[name] || instance.$options[name])
                    opts[name] = args[name];
                else
                    throw new error.InvalidOption("Invalid option passed: " + name);
            });
        }
        
        return [new (instance)(opts), method];
    };
    
    var globalOptions = {
        "app": true
    };
    
    this.run = function(cmd, args, callback) {
        if (_.isFunction(args)) {
            callback = args;
            args = null;
        }
        var ret = this.prepareRun(cmd, args);
        var object = ret[0];
        var method = ret[1];
        
        this.currentCommand = cmd;
        object[method].call(object, callback);
    };
    
}).call(HerokuCommand.prototype);