"use strict";

var Sys = require("sys");

/**
 * JavaScript native error types:
 *  - EvalError: An error in the eval() function has occurred
 *  - RangeError: Out of range number value has occurred
 *  - ReferenceError: An illegal reference has occurred
 *  - SyntaxError: A syntax error within code inside the eval() function has occurred.
 *                 All other syntax errors are not caught by try/catch/finally, and will
 *                 trigger the default engine error message associated with the error.
 *                 To catch actual syntax errors, you may use the onerror event.
 *  - TypeError: An error in the expected variable type has occurred
 *  - URIError: An error when encoding or decoding the URI has occurred (ie: when calling encodeURI()).
 */

/**
 * Exception thrown if an error which can only be found on runtime occurs.
 */
var RuntimeError = function(message, code) {
    Error.call(this);
    this.name = this.type = "Runtime";
    this.code = code || "error"; // error | warn | notice
    this.message = message || this.type;
};
Sys.inherits(RuntimeError, Error);

/**
 * Exception that represents error in the program logic.
 * This kind of exceptions should directly lead to a fix in your code.
 */
var LogicError = function(message, code) {
    Error.call(this);
    this.name = this.type = "Logic";
    this.code = code || "error"; // error | warn | notice
    this.message = message || this.type;
};
Sys.inherits(LogicError, Error);

/**
 *  Exception thrown when the expected variable type is not correct.
 */
var TypeError = exports.Type = function(message, code) {
    LogicError.call(this, message, code);
    this.name = this.type = "Type";
};
Sys.inherits(TypeError, LogicError);

/**
 * Extending Error for our own custom error types
 */
var InvalidArgumentError = exports.InvalidArgument = function(message, code) {
    TypeError.call(this, message, code);
    this.name = "InvalidArgument";
};
Sys.inherits(InvalidArgumentError, LogicError);

var InvalidOptionError = exports.InvalidOption = function(message, code) {
    InvalidArgumentError.call(this, message, code);
    this.name = "InvalidOption";
};
Sys.inherits(InvalidOptionError, InvalidArgumentError);

var IOError = exports.IO = function(message, code) {
    RuntimeError.call(this, message, code);
    this.name = "IO";
};
Sys.inherits(IOError, RuntimeError);

var BoundsError = exports.Bounds = function(message, code) {
    RuntimeError.call(this, message, code);
    this.name = "Bounds";
};
Sys.inherits(BoundsError, RuntimeError);

var BadFunctionCallError = exports.BadFunctionCall = function(message, code) {
    LogicError.call(this, message, code);
    this.name = "BadFunctionCall";
};
Sys.inherits(BadFunctionCallError, LogicError);

/**
 * HttpError types
 */
var HttpError = exports.Http = function(message, code) {
    IOError.call(this, message, code);
    this.name = this.type = "Http";
    //Error.captureStackTrace(this, arguments.callee);
};
Sys.inherits(HttpError, IOError);

HttpError.prototype.toString = function() {
    return this.message;
};

var statusCodes = {
    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Request Entity Too Large",
    414: "Request-URI Too Long",
    415: "Unsupported Media Type",
    416: "Requested Range Not Satisfiable",
    417: "Expectation Failed",
    422: "Unprocessable Entity",
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Timeout",
    505: "HTTP Version Not Supported"
};

var status, defaultMsg, className, error;

for (status in statusCodes) {
    defaultMsg = statusCodes[status];
    className = toCamelCase(defaultMsg);
    
    error = (function(defaultMsg, status, className) {
        return function(msg) {
            HttpError.call(this, msg || status + ": " + defaultMsg, Number(status));
            this.name = className;
            
            if (status >= 500)
                Error.captureStackTrace(this, arguments.callee);
        };
    })(defaultMsg, status, className);
    
    Sys.inherits(error, HttpError);
    
    exports[className] = error;
    exports[status] = error;
}

function toCamelCase(str) {
    return str.toLowerCase().replace(/(?:(^.)|(\s+.))/g, function(match) {
        return match.charAt(match.length-1).toUpperCase();
    });
}

/**
 * Extending Http errors for Validation
 */
var ValidationError = exports.Validation = function(object) {
    exports[400].call(this, object.message, 400);
    this.name = "Validation";
    this.instance = object;
};
Sys.inherits(ValidationError, HttpError);

ValidationError.prototype.toJSON = function() {
    return this.instance;
};
